from ctypes import c_uint8

from spc_io.misc import EnumType, Union


class Fxtype(metaclass=EnumType):
    _type_ = c_uint8
    _enums_ = {
        0: 'XARB',        # Arbitrary
        1: 'XWAVEN',      # Wavenumber (cm-1)
        2: 'XUMETR',      # Micrometers (um)
        3: 'XNMETR',      # Nanometers (nm)
        4: 'XSECS',       # Seconds
        5: 'XMINUTS',     # Minutes
        6: 'XHERTZ',      # Hertz (Hz)
        7: 'XKHERTZ',     # Kilohertz (KHz)
        8: 'XMHERTZ',     # Megahertz (MHz)
        9: 'XMUNITS',     # Mass (M/z)
        10: 'XPPM',       # Parts per million (PPM)
        11: 'XDAYS',      # Days
        12: 'XYEARS',     # Years
        13: 'XRAMANS',    # Raman Shift (cm-1)
        14: 'XEV',        # eV
        15: 'ZTEXTL',     # XYZ text labels in fcatxt (old 0x4D version only)
        16: 'XDIODE',     # Diode Number
        17: 'XCHANL',     # Channel
        18: 'XDEGRS',     # Degrees
        19: 'XDEGRF',     # Temperature (F)
        20: 'XDEGRC',     # Temperature (C)
        21: 'XDEGRK',     # Temperature (K)
        22: 'XPOINT',     # Data Points
        23: 'XMSEC',      # Milliseconds (mSec)
        24: 'XUSEC',      # Microseconds (uSec)
        25: 'XNSEC',      # Nanoseconds (nSec)
        26: 'XGHERTZ',    # Gigahertz (GHz)
        27: 'XCM',        # Centimeters (cm)
        28: 'XMETERS',    # Meters (m)
        29: 'XMMETR',     # Millimeters (mm)
        30: 'XHOURS',     # Hours
        255: 'XDBLIGM',   # Double interferogram (no display labels)
    }


class Fytype(metaclass=EnumType):
    _type_ = c_uint8
    _enums_ = {
        0: 'YARB',        # Arbitrary Intensity
        1: 'YIGRAM',      # Interferogram
        2: 'YABSRB',      # Absorbance
        3: 'YKMONK',      # Kubelka-Monk
        4: 'YCOUNT',      # Counts
        5: 'YVOLTS',      # Volts
        6: 'YDEGRS',      # Degrees
        7: 'YAMPS',       # Milliamps
        8: 'YMETERS',     # Millimeters
        9: 'YMVOLTS',     # Millivolts
        10: 'YLOGDR',     # Log(1/R)
        11: 'YPERCNT',    # Percent
        12: 'YINTENS',    # Intensity
        13: 'YRELINT',    # Relative Intensity
        14: 'YENERGY',    # Energy
        16: 'YDECBL',     # Decibel
        19: 'YDEGRF',     # Temperature (F)
        20: 'YDEGRC',     # Temperature (C)
        21: 'YDEGRK',     # Temperature (K)
        22: 'YINDRF',     # Index of Refraction [N]
        23: 'YEXTCF',     # Extinction Coeff. [K]
        24: 'YREAL',      # Real
        25: 'YIMAG',      # Imaginary
        26: 'YCMPLX',     # Complex
        128: 'YTRANS',    # Transmission (ALL TYPES >= 128 ARE ASSUMED INVERTED PEAKS!)
        129: 'YREFLEC',   # Reflectance
        130: 'YVALLEY',   # Arbitrary or Single Beam with Valley Peaks
        131: 'YEMISN',    # Emission
        255: 'YREFARBE',  # Reference Arbitrary Energy
    }


class Fxytype(Union):
    _pack_ = 1
    _anonymous_ = ('fxtype', 'fytype')
    _fields_ = [
        ('fxtype', Fxtype),
        ('fytype', Fytype),
    ]



from ctypes import addressof, c_char, c_uint8, c_uint32, sizeof, string_at
from typing import Dict, Union

from pydantic import validate_call
from spc_io.misc import Structure


class LogBookBase:
    def txt_as_dict(self):
        return dict([i.decode(errors='surrogateescape').split('=')
                     for ii in self.txt.split(b'\r\n')
                     for i in ii.split(b'\n\r')
                     if b'=' in i
                     ])

    def disk_as_bytes(self):
        return bytearray(string_at(addressof(self.disk), sizeof(self.disk)))

    def binary_as_bytes(self):
        return bytearray(string_at(addressof(self.binary), sizeof(self.binary)))


class Logstc(Structure):
    _pack_ = 1
    _fields_ = [
        ('logsizd', c_uint32),     # byte size of disk block
        ('logsizm', c_uint32),     # byte size of memory block
        ('logtxto', c_uint32),     # byte offset to text
        ('logbins', c_uint32),     # byte size of binary area (immediately after logstc)
        ('logdsks', c_uint32),     # byte size of disk area (immediately after logbins)
        ('logspar', c_char*44),    # reserved (must be zero)
    ]

    @classmethod
    @validate_call
    def new_header_and_logbook_from_data(cls, *,
                                         disk: bytes = b'',
                                         binary: bytes = b'',
                                         txt: Union[bytes, Dict] = b''):
        if isinstance(txt, dict):
            txt = b'\r\n'.join([f'{k}={v}'.encode(errors='surrogateescape') for k, v in txt.items()])
        txt += b'\x00'  # NUL terminate
        logsizd = len(disk)+len(binary)+len(txt)+sizeof(cls)
        logsizm = ((logsizd // 4096) + 1) * 4096
        logbins = len(binary)
        logdsks = len(disk)
        logtxto = (sizeof(cls)+logbins+logdsks)
        ret = cls(logsizd=logsizd,
                  logsizm=logsizm,
                  logtxto=logtxto,
                  logbins=logbins,
                  logdsks=logdsks,
                  logspar=b'\x00',
                  )
        LogBook = ret.build_LogBook_type()
        log_book = LogBook.from_buffer_copy(disk + binary + txt)
        return ret, log_book

    def build_LogBook_type(self):
        dct = {
            '_pack_': 1,
            '_fields_': [
                ('disk', c_uint8 * self.logdsks),
                ('binary', c_uint8 * self.logbins),
                ('_skip', c_uint8 * (self.logtxto - sizeof(self) - (self.logdsks + self.logbins))),
                ('txt', c_char * (self.logsizd - self.logtxto)),
            ]
        }
        return type('LogBook', (Structure, LogBookBase), dct)



from numpy import linspace
from pydantic import validate_call


class EvenAxis:
    @validate_call
    def __init__(self, start: float, stop: float, num: int):
        self._start = start
        self._stop = stop
        self._num = num
        self._values_array = None

    def __repr__(self):
        return f'{type(self).__name__}({self.start}, {self.stop}, {self.num})'

    def __len__(self):
        return self.num

    @property
    def values(self):
        if self._values_array is None:
            self._values_array = linspace(start=self.start, stop=self.stop, num=self.num)
        return self._values_array

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @property
    def num(self):
        return self._num



import logging
from ctypes import Array, c_float, c_int16, c_int32
from typing import Literal, Union

import numpy as np
import numpy.typing as npt
from pydantic import validate_call

from .headers.sub_hdr import SubHdr
from .xarray_property import XArrayProperty

logger = logging.getLogger(__name__)


class SubFile(XArrayProperty):
    @validate_call(config=dict(arbitrary_types_allowed=True))
    def __init__(self, header: SubHdr,
                 xarray: Union[npt.NDArray, Array, None] = None,
                 yarray: Union[npt.NDArray, Array, None] = None,
                 single_prec: bool = False,
                 w=False, z=False):

        # Meaning of values for w and z
        # False - the axis is disabled
        # True - the value is calcualted usinc boundaries and increment
        # or a number for abitrary values

        self.header = header

        if self.header.subexp == -0x80 or isinstance(yarray, Array) and yarray._type_ == c_float:
            self._yarray_type = c_float
        else:
            if single_prec:
                self._yarray_type = c_int16
            else:
                self._yarray_type = c_int32

        self.xarray = xarray
        self.yarray = yarray

        self._w = w
        self._z = z

    @property
    def yarray_type(self):
        return self._yarray_type

    @yarray_type.setter
    @validate_call(config=dict(arbitrary_types_allowed=True))
    def yarray_type(self, val: Literal[c_float, c_int16, c_int32]):
        self._yarray_type = val

    @property
    def z(self):
        if self._z is True:
            return self.header.subfirst
        elif self._z is False:
            return None
        else:
            return self._z

    @z.setter
    def z(self, val):
        # support only non evenly distributed z
        self._z = True
        self.header.subfirst = val

    @z.deleter
    def z(self):
        self._z = False

    @property
    def w(self):
        if self._w is True:
            return self.header.subwlevel
        elif self._w is False:
            return None
        else:
            return self._w

    @w.setter
    def w(self, val):
        # support only non evenly distributed w
        self._w = True
        self.header.subwlevel = val

    @w.deleter
    def w(self):
        self._w = False

    @property
    def yarray(self):
        arr = np.ctypeslib.as_array(self._yarray)
        if self._yarray_type == c_float:
            return arr
        elif self._yarray_type == c_int16:
            return (2. ** (self.header.subexp-16)) * arr
        elif self._yarray_type == c_int32:
            return (2. ** (self.header.subexp-32)) * arr

    @yarray.setter
    @validate_call(config=dict(arbitrary_types_allowed=True))
    def yarray(self, arr: Union[npt.NDArray, Array, None]):
        if arr is None:
            self._yarray = arr
        elif isinstance(arr, Array):
            if arr._type_ != self._yarray_type:
                raise ValueError(f'wrong array type. Provided {arr._type_}, expected {self._yarray_type}')
            self._yarray = arr
        else:
            if self._yarray_type == c_int16:
                arr = arr / (2. ** (self.header.subexp-16))
            elif self._yarray_type == c_int32:
                arr = arr / (2. ** (self.header.subexp-32))
            self._yarray = np.ctypeslib.as_ctypes(arr.astype(self._yarray_type))



from ctypes import c_uint8

from spc_io.misc import EnumType


class Fexper(metaclass=EnumType):
    _type_ = c_uint8
    _enums_ = {
        0: 'SPCGEN',    # General SPC (could be anything)
        1: 'SPCGC',     # Gas Chromatogram
        2: 'SPCCGM',    # General Chromatogram (same as SPCGEN with TCGRAM)
        3: 'SPCHPLC',   # HPLC Chromatogram
        4: 'SPCFTIR',   # FT-IR, FT-NIR, FT-Raman Spectrum or Igram (Can also be used for scanning IR.)
        5: 'SPCNIR',    # NIR Spectrum (Usually multi-spectral data sets for calibration.)
        7: 'SPCUV',     # UV-VIS Spectrum (Can be used for single scanning UV-VIS-NIR)
        8: 'SPCXRY',    # X-ray Diffraction Spectrum
        9: 'SPCMS',     # Mass Spectrum (Can be single, GC-MS, Continuum, Centroid or TOF.)
        10: 'SPCNMR',   # NMR Spectrum or FID
        11: 'SPCRMN',   # Raman Spectrum (Usually Diode Array, CCD, etc. use SPCFTIR for FT-Raman.)
        12: 'SPCFLR',   # Fluorescence Spectrum
        13: 'SPCATM',   # Atomic Spectrum
        14: 'SPCDAD',   # Chromatography Diode Array Spectra
    }



from ctypes import LittleEndianStructure


class EnumMethods(LittleEndianStructure):
    @property
    def UNKNOWN(self):
        return self._val not in self._enums_

    def __init__(self, _val):
        if isinstance(_val, str):
            num = list(self._enums_.keys())[list(self._enums_.values()).index(_val)]
            super().__init__(_val=num)
        else:
            super().__init__(_val=_val)

    def __str__(self):
        if self._val not in self._enums_:
            return 'UNKNOWN'
        else:
            return self._enums_[self._val]

    def __repr__(self):
        return f'{type(self).__name__}({self._val})'


class EnumType(type):
    def __new__(cls, name, bases, dct):
        dct_struct = dict(
            _pack_=1,
            _fields_=[('_val', dct['_type_'])],
            )
        dct_struct.update(dct)
        dct_struct.update({
            enum: property(lambda self, enumt=(enum,):
                           self._val in dct['_enums_'] and dct['_enums_'][self._val] == enumt[0])
            for enum in dct['_enums_'].values()
        })
        ret = type(name, (EnumMethods,), dct_struct)
        return ret



from datetime import datetime
from typing import Dict, Union

from pydantic import validate_call


class LogBook:
    @validate_call
    def __init__(self, *,
                 disk: bytes = b'',
                 binary: bytes = b'',
                 text: Dict[str, Union[int, float, datetime, str]] = dict()):
        self.disk = disk
        self.binary = binary
        self.text = text



from ctypes import c_uint8

from spc_io.misc import Structure


class Ftflgs(Structure):
    _pack_ = 1
    _fields_ = [
        ('TSPREC', c_uint8, 1),   # Single precision (16 bit) Y data if set.
        ('TCGRAM', c_uint8, 1),   # Enables fexper in older software (not used)
        ('TMULTI', c_uint8, 1),   # Multiple traces format (set if more than one subfile)
        ('TRANDM', c_uint8, 1),   # If TMULTI and TRANDM=1 then arbitrary time (Z) values
        ('TORDRD', c_uint8, 1),   # If TMULTI abd TORDRD=1 then ordered but uneven subtimes
        ('TALABS', c_uint8, 1),   # Set if should use fcatxt axis labels, not fxtype etc.
        ('TXYXYS', c_uint8, 1),   # If TXVALS and multifile, then each subfile has own X's
        ('TXVALS', c_uint8, 1),   # Floating X value array preceeds Y's (New format only)
    ]



import ctypes


class ComplexTypeBase:
    def __str__(self):
        kw = ', '.join([f'{k}={str(v)}' for k, v in self.to_dict().items() if not k.startswith('_')])
        return f'{type(self).__name__}({kw})'

    def __repr__(self):
        return str(self)

    def to_dict(self):
        return {field: str(getattr(self, field)) for field, *_ in self._fields_}


class Structure(ctypes.LittleEndianStructure, ComplexTypeBase):
    pass


class Union(ctypes.Union, ComplexTypeBase):
    pass



from ctypes import c_float, c_uint32

from spc_io.misc import Structure


class Ssfstc(Structure):
    _pack_ = 1
    _fields_ = [
        ('ssfposn', c_uint32),   # disk file position of beginning of subfile (subhdr)
        ('ssfsize', c_uint32),   # byte size of subfile (subhdr+X+Y)
        ('ssftime', c_float),    # floating Z time of subfile (subtime)
    ]



import io
import logging
from ctypes import addressof, c_float, c_int16, c_int32, sizeof, string_at
from typing import List, Union

from pydantic import validate_call

from .headers.logstc import Logstc
from .headers.spchdr import SpcHdr
from .headers.ssfstc import Ssfstc
from .headers.sub_hdr import SubHdr
from .sub_file import SubFile
from .xarray_property import XArrayProperty

logger = logging.getLogger(__name__)


class SpcRaw(XArrayProperty):
    @classmethod
    @validate_call(config=dict(arbitrary_types_allowed=True))
    def from_high_level(cls, *,
                        header: SpcHdr,
                        subs: List[SubFile],
                        xarray=None,
                        log_binary: bytes = b'',
                        log_disk: bytes = b'',
                        log_txt={},
                        ):
        self = cls()
        self.main_header = header
        self.xarray = xarray
        self.subs = subs
        self.log_header, self.log_book = Logstc.new_header_and_logbook_from_data(
            disk=log_disk, binary=log_binary, txt=log_txt)
        self.dirs = list()
        self.calcuate_offsets()
        return self

    @classmethod
    def from_bytes_io(cls, bytes_io: Union[io.BytesIO, io.BufferedReader]):
        self = cls()
        self.subs = list()
        self._xarray = None
        self.log_header = None
        self.log_book = None
        self.dirs = list()

        def read_bytes(obj_type):
            nbytes = sizeof(obj_type)
            return obj_type.from_buffer_copy(bytes_io.read(nbytes))

        self.main_header = read_bytes(SpcHdr)
        if self.main_header.ftflgs.TRANDM:
            raise NotImplementedError('ftflgs.TRANDM bit is not supproted')

        if not self.main_header.fversn.LSB_format:
            raise NotImplementedError('only LSB format is supproted')

        # global X
        if self.main_header.ftflgs.TXVALS and not self.main_header.ftflgs.TXYXYS:
            # single X array for all Y
            self._xarray = read_bytes(c_float*self.main_header.fnpts)

        # subfiles
        for sub_i in range(self.main_header.fnsub):
            sub_header = read_bytes(SubHdr)
            if self.main_header.fexp == -0x80 or sub_header.subexp == -0x80:
                ydata_type = c_float
            else:
                if self.main_header.ftflgs.TSPREC:  # single precision
                    ydata_type = c_int16
                else:
                    ydata_type = c_int32
            if self.main_header.ftflgs.TXYXYS:
                # particular x
                xarray = read_bytes(c_float*sub_header.subnpts)
                yarray = read_bytes(ydata_type*sub_header.subnpts)
            else:
                xarray = None
                yarray = read_bytes(ydata_type*self.main_header.fnpts)
            sub_file = SubFile(sub_header,
                               xarray=xarray,
                               yarray=yarray,
                               single_prec=self.main_header.ftflgs.TSPREC)
            self.subs.append(sub_file)

        # fix w axis
        w_first = self.subs[0].header.subwlevel
        for sub in self.subs:
            if self.main_header.fwplanes:
                # w axis enabled
                if self.main_header.fwinc:
                    # w values are evenly distributed
                    w_i = sub.header.subindx // self.main_header.fwplanes
                    sub._w = w_first + self.main_header.fwinc * w_i
                else:
                    sub._w = True  # take from subfile header
            else:
                sub._w = False  # W axis is disabled

        # fix z axis
        if self.main_header.ftflgs.TMULTI:
            if self.main_header.fzinc == 0:
                self.main_header.fzinc = self.subs[0].header.subnext - self.subs[0].header.subfirst
            if not self.main_header.ftflgs.TORDRD or self.main_header.ftflgs.TRANDM:
                first = self.subs[0].header.subfirst
                increment = self.main_header.fzinc
                for sub in self.subs:
                    if self.main_header.fwplanes:
                        sub._z = first + increment * (sub.header.subindx % self.main_header.fwplanes)
                    else:
                        sub._z = first + increment * sub.header.subindx
            else:
                for sub in self.subs:
                    sub._z = True  # take from subfile header
        else:
            sub._z = False  # Z axis is disabled

        # directory
        if self.main_header.ftflgs.TXYXYS:
            if self.main_header.fnpts:
                # assume packed
                # bytes_io.seek(self.main_header.fnpts)
                for sub_i in range(self.main_header.fnsub):
                    self.dirs.append(read_bytes(Ssfstc))

        # Log book
        if self.main_header.flogoff:
            # assume packed
            # bytes_io.seek(self.main_header.flogoff)
            self.log_header = read_bytes(Logstc)
            LogBook = self.log_header.build_LogBook_type()
            self.log_book = read_bytes(LogBook)
        else:
            self.log_header, self.log_book = Logstc.new_header_and_logbook_from_data()

        self._extra_bytes_at_eof = bytes_io.read()
        if len(self._extra_bytes_at_eof):
            logger.warning(f'{len(self._extra_bytes_at_eof)} left at the end of the file')
        return self

    def calcuate_offsets(self):
        self.dirs = list()
        ofs = sizeof(self.main_header)
        if self.main_header.ftflgs.TXVALS and not self.main_header.ftflgs.TXYXYS:
            ofs += sizeof(self._xarray)
        for sub in self.subs:
            size = sizeof(sub.header)
            if self.main_header.ftflgs.TXVALS and self.main_header.ftflgs.TXYXYS:
                size += sizeof(sub._xarray)
            size += sizeof(sub._yarray)
            ofs += size
            if sub.z is not None:
                self.dirs.append(Ssfstc(ssfposn=ofs, ssfsize=size, ssftime=sub.z))

        # Directory offset
        ofs += sizeof(self.log_book)
        if self.main_header.ftflgs.TXYXYS:
            self.main_header.fnpts = ofs

        # Logbook offset
        self.main_header.flogoff = ofs

        self.main_header.fnsub = len(self.subs)

    def _to_bytes_as_is(self) -> bytes:
        def to_bytes(c_data):
            return string_at(addressof(c_data), sizeof(c_data))
        ret = []

        # Main header
        ret.append(to_bytes(self.main_header))

        # Global X values
        if self._xarray is not None:
            ret.append(to_bytes(self._xarray))

        # Subfiles
        for sub in self.subs:
            ret.append(to_bytes(sub.header))
            if sub._xarray is not None:
                ret.append(to_bytes(sub._xarray))
            ret.append(to_bytes(sub._yarray))

        # Directories
        for dd in self.dirs:
            ret.append(to_bytes(dd))

        # Logbook
        if self.log_header.logsizd > 1:  # 1 because of NUL terminated string
            ret.append(to_bytes(self.log_header))
            ret.append(to_bytes(self.log_book))
        return b''.join(ret)

    def to_bytes(self) -> bytes:
        self.calcuate_offsets()
        return self._to_bytes_as_is()



from ctypes import (c_char, c_double, c_float, c_int8, c_uint8, c_uint16,
                    c_uint32)

from spc_io.misc import Structure

from .fdate import Fdate
from .fexper import Fexper
from .fmods import Fmods
from .ftflgs import Ftflgs
from .fversn import Fversn
from .fxytype import Fxtype, Fytype


class SpcHdr(Structure):
    _pack_ = 1
    _anonymous_ = ()
    _fields_ = [
        ('ftflgs', Ftflgs),         # Flag bits defined below
        ('fversn', Fversn),         # 0x4B=> new LSB 1st, 0x4C=> new MSB 1st, 0x4D=> old format
        ('fexper', Fexper),         # Instrument technique code (see below)
        ('fexp', c_int8),           # Fraction scaling exponent integer (80h=>float)
        ('fnpts', c_uint32),        # Integer number of points (or TXYXYS directory position
        ('ffirst', c_double),       # Floating X coordinate of first point
        ('flast', c_double),        # Floating X coordinate of last point
        ('fnsub', c_uint32),        # Integer number of subfiles (1 if not TMULTI)
        ('fxtype', Fxtype),         # Type of X axis units (see definitions below)
        ('fytype', Fytype),         # Type of Y axis units (see definitions below)
        ('fztype', Fxtype),         # Type of Z axis units (see definitions below)
        ('fpost', c_uint8),         # Posting disposition (see GRAMSDDE.H)
        ('fdate', Fdate),           # Date/Time LSB: min=6b,hour=5b,day=5b,month=4b,year=12b
        ('fres', c_char*9),         # Resolution description text (null terminated)
        ('fsource', c_char*9),      # Source instrument description text (null terminated)
        ('fpeakpt', c_uint16),      # Peak point number for interferograms (0=not known)
        ('fspare', c_float*8),      # Used for Array Basic storage
        ('fcmnt', c_char*130),      # Null terminated comment ASCII text string
        ('fcatxt', c_char*30),      # X,Y,Z axis label strings if ftflgs=TALABS
        ('flogoff', c_uint32),      # File offset to log block or 0 (see above)
        ('fmods', Fmods),           # File Modification Flags (see below: 1=A,2=B,4=C,8=D..)
        ('fprocs', c_uint8),        # Processing code (see GRAMSDDE.H)
        ('flevel', c_uint8),        # Calibration level plus one (1 = not calibration data)
        ('fsampin', c_uint16),      # Sub-method sample injection number (1 = first or only )
        ('ffactor', c_float),       # Floating data multiplier concentration factor (IEEE-32)
        ('fmethod', c_char*48),     # Method/program/data filename w/extensions comma list
        ('fzinc', c_float),         # Z subfile increment (0 = use 1st subnext-subfirst)
        ('fwplanes', c_uint32),     # Number of planes for 4D with W dimension (0=normal)
        ('fwinc', c_float),         # W plane increment (only if fwplanes is not 0)
        ('fwtype', Fxtype),         # Type of W axis units (see definitions below)
        ('freserv', c_char*187),    # Reserved (must be set to zero)
    ]
